#
# Author: Manuel de la Pena <manuel@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Specific tests for the os_helper on Windows."""

import os
import errno

from ubuntuone.platform.windows.os_helper import (
    remove_windows_illegal_chars,
    access,
    is_illegal_path,
    listdir,
    make_dir,
    normpath,
    open_file,
    path_exists,
    remove_dir,
    remove_file,
    rename,
    remove_utf8_chars,
    set_no_rights,
    set_file_readwrite,
    stat_path,
    LONG_PATH_PREFIX,
    WINDOWS_ILLEGAL_CHARS_MAP
)
from contrib.testing.testcase import BaseTwistedTestCase


class TestIllegalPaths(BaseTwistedTestCase):
    """Test all the operations using illegal paths."""
    
    def setUp(self):
        """Setup for the tests."""
        super(TestIllegalPaths, self).setUp()
        self.basedir = self.mktemp('test_root')
        self.testfile = os.path.join(self.basedir, "test_file")
        self.testfile += ''.join(WINDOWS_ILLEGAL_CHARS_MAP)
        self.fd = open(remove_windows_illegal_chars(self.testfile), 'w')

    def test_path_file_exist_yes(self):
        """Test that the file exists."""
        self.assertTrue(path_exists(self.testfile))

    def test_path_file_exist_no(self):
        """Test that the file doesn't exist."""
        self.fd.close()
        os.remove(remove_windows_illegal_chars(self.testfile))
        self.assertFalse(path_exists(self.testfile))

    def test_path_dir_exist_yes(self):
        """Test that the dir exists."""
        self.assertTrue(path_exists(self.basedir))

    def test_path_dir_exist_no(self):
        """Test that the dir doesn't exist."""
        self.fd.close()
        os.remove(remove_windows_illegal_chars(self.testfile))
        self.assertFalse(path_exists(os.path.join(self.basedir, 'nodir')))

    def test_remove_file(self):
        """Test the remove file."""
        self.fd.close()
        remove_file(self.testfile)
        self.assertFalse(os.path.exists(
                         remove_windows_illegal_chars(self.testfile)))

    def test_remove_dir(self):
        """Test the remove dir."""
        testdir = os.path.join(self.basedir, 'foodir')
        os.mkdir(testdir)
        assert os.path.exists(testdir)
        remove_dir(testdir)
        self.assertFalse(path_exists(testdir))

    def test_make_dir_one(self):
        """Test the make dir with one dir."""
        testdir = os.path.join(self.basedir, 'foodir')
        assert not os.path.exists(testdir)
        make_dir(testdir)
        self.assertTrue(os.path.exists(testdir))

    def test_make_dir_already_there(self):
        """Test the make dir with one dir that exists."""
        self.assertRaises(OSError, make_dir, self.basedir)

    def test_make_dir_recursive_no(self):
        """Test the make dir with some dirs, not recursive explicit."""
        testdir = os.path.join(self.basedir, 'foo', 'bar')
        assert not os.path.exists(testdir)
        self.assertRaises(OSError, make_dir, testdir)

    def test_make_dir_recursive_yes(self):
        """Test the make dir with some dirs, recursive."""
        testdir = os.path.join(self.basedir, 'foo', 'bar')
        assert not os.path.exists(testdir)
        make_dir(testdir, recursive=True)
        self.assertTrue(os.path.exists(testdir))

    def test_open_file_not_there(self):
        """Open a file that does not exist."""
        self.assertRaises(IOError, open_file, os.path.join(self.basedir, 'no'))

    def test_open_file_gets_a_fileobject(self):
        """Open a file, and get a file object."""
        testfile = os.path.join(self.basedir, 'testfile')
        open(testfile, 'w').close()
        f = open_file(testfile)
        self.assertTrue(isinstance(f, file))

    def test_open_file_read(self):
        """Open a file, and read."""
        testfile = os.path.join(self.basedir, 'testfile')
        with open(testfile, 'w') as fh:
            fh.write("foo")
        f = open_file(testfile, 'r')
        self.assertTrue(f.read(), "foo")

    def test_open_file_write(self):
        """Open a file, and write."""
        testfile = os.path.join(self.basedir, 'testfile')
        with open_file(testfile, 'w') as fh:
            fh.write("foo")
        f = open(testfile)
        self.assertTrue(f.read(), "foo")

    def test_rename_not_there(self):
        """Rename something that does not exist."""
        self.assertRaises(OSError, rename,
                          os.path.join(self.basedir, 'no>'), 'foo')

    def test_rename_file(self):
        """Rename a file."""
        testfile1 = os.path.join(self.basedir, 'testfile1>')
        testfile2 = os.path.join(self.basedir, 'testfile2<')
        open(remove_windows_illegal_chars(testfile1), 'w').close()
        rename(testfile1, testfile2)
        self.assertFalse(os.path.exists(
                         remove_windows_illegal_chars(testfile1)))
        self.assertTrue(os.path.exists(
                         remove_windows_illegal_chars(testfile2)))

    def test_rename_dir(self):
        """Rename a dir."""
        testdir1 = os.path.join(self.basedir, 'testdir1>')
        testdir2 = os.path.join(self.basedir, 'testdir2<')
        os.mkdir(remove_windows_illegal_chars(testdir1))
        rename(testdir1, testdir2)
        self.assertFalse(os.path.exists(
                         remove_windows_illegal_chars(testdir1)))
        self.assertTrue(os.path.exists(
                         remove_windows_illegal_chars(testdir2)))

    def test_listdir(self):
        """Return a list of the files in a dir."""
        open(os.path.join(self.basedir, 'foo'), 'w').close()
        open(os.path.join(self.basedir, 'bar'), 'w').close()
        l = listdir(self.basedir)
        test_file = u'test_file\u200b\u2033\u200b\u200b\u203d\u200b\u200b' +\
                    u'\u26b9\u200b\u200b\u2223\u200b\u200b\u2236\u200b\u200b' +\
                    u'\u2039\u200b\u200b\u2044\u200b\u200b\u203a\u200b'
        self.assertEqual(sorted(l), ['bar', 'foo', test_file])

    def test_access_rw(self):
        """Test access on a file with full permission."""
        self.assertTrue(access(self.testfile))

    def test_access_ro(self):
        """Test access on a file with read only permission."""
        fixed_path = remove_windows_illegal_chars(self.testfile)
        os.chmod(fixed_path, 0o444)
        self.assertTrue(access(self.testfile))
        os.chmod(fixed_path, 0o664)

    def test_access_nothing(self):
        """Test access on a file with no permission at all."""
        set_no_rights(self.testfile)
        self.assertFalse(access(self.testfile))
        set_file_readwrite(self.testfile)

    def test_stat_normal(self):
        """Test on a normal file."""
        self.assertEqual(os.stat(remove_windows_illegal_chars(self.testfile)),
                         stat_path(self.testfile))

    def test_stat_no_path(self):
        """Test that it raises proper error when no file is there."""
        try:
            return stat_path(os.path.join(self.basedir, 'nofile'))
        except OSError, e:
            self.assertEqual(e.errno, errno.ENOENT)

    def test_path_exists_file_yes(self):
        """The file is there."""
        self.assertTrue(path_exists(self.testfile))

    def test_path_exists_file_no(self):
        """The file is not there."""
        self.fd.close()
        os.remove(remove_windows_illegal_chars(self.testfile))
        self.assertFalse(path_exists(self.testfile))

    def test_path_exists_dir_yes(self):
        """The dir is there."""
        self.assertTrue(path_exists(self.basedir))

    def test_path_exists_dir_no(self):
        """The dir is not there."""
        self.fd.close()
        self.assertFalse(path_exists(os.path.join(self.basedir, 'subdir')))

    def test_remove_utf8_chars(self):
        """Test that the path does not have utf chars."""
        path = remove_utf8_chars(remove_windows_illegal_chars(self.testfile))
        self.assertFalse(is_illegal_path(path))


class TestSpecialOSCalls(BaseTwistedTestCase):
    """Test those calls that have extra logic."""

    def test_normpath_with_longprefix(self):
        """Ensure that the normpath is correct when it uses the long prefix"""
        paths = [os.path.join('A','B'),
                 os.path.join('A', 'B') + os.path.sep,
                 os.path.join('A', '.', 'B'),
                 os.path.join('A', 'foo', '..', 'B')]
        for current_path in paths:
            long_path = LONG_PATH_PREFIX + current_path
            self.assertEqual(LONG_PATH_PREFIX + os.path.normpath(current_path),
                             normpath(long_path))