# tests.platform.windows.test_vm_helper - windows vm helper tests
#
# Authors: Guillermo Gonzalez <guillermo.gonzalez@canonical.com>
#          Manuel de la Pena <manuel@canonical.com>
# Copyright 2010 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Windows specific tests for vm_helper."""

import os
import uuid
from mocker import Mocker

from contrib.testing.testcase import environ
from tests.syncdaemon.test_vm import BaseVolumeManagerTests

from ubuntuone.platform.windows.os_helper import LONG_PATH_PREFIX
from ubuntuone.platform.windows.vm_helper import (
    create_shares_link, get_udf_path, get_udf_path_name, get_share_path)


class VMHelperTest(BaseVolumeManagerTests):
    """Test the vm_helper methods."""

    def setUp(self):
        """Set the different tests."""
        super(VMHelperTest, self).setUp()
        self.mocker = Mocker()

    def test_create_shares_link(self):
        """Test for create_shares_link."""
        source = 'source'
        dest = 'dest'
        make_link = self.mocker.replace(
                            'ubuntuone.platform.windows.os_helper.make_link')
        make_link(source, dest)
        self.mocker.replay()
        create_shares_link(source, dest)
        self.mocker.verify()

    def test_get_udf_path_name(self):
        """Test for _get_udf_path_name."""
        home_dir = 'C:\\Users\\ubuntuonehacker'
        in_home = os.path.join(home_dir, 'foo')
        deep_in_home = os.path.join(home_dir, 'docs', 'foo', 'bar')
        outside_home = os.path.join(os.path.sep, 'home', 'other', 'foo')
        relative_home = '..\\..\\home\\foo\\bar'
        with environ('HOME', home_dir):
            path, name = get_udf_path_name(in_home)
            self.assertEquals('~', path)
            self.assertEquals('foo', name)
            path, name = get_udf_path_name(deep_in_home)
            self.assertEquals('~\\docs\\foo', path)
            self.assertEquals('bar', name)
            self.assertRaises(ValueError, get_udf_path_name, outside_home)
            self.assertRaises(ValueError, get_udf_path_name, None)
            self.assertRaises(ValueError, get_udf_path_name, relative_home)

    def test_get_udf_path(self):
        """Test that we do get a correct path."""
        path = 'path/to/udf'
        os_path = self.mocker.replace('os.path')
        os_path.expanduser(path.replace('/', '\\'))
        self.mocker.result('path')
        self.mocker.replay()
        udf_path = get_udf_path(path) 
        self.assertTrue(LONG_PATH_PREFIX in udf_path)
        self.assertFalse('/' in udf_path)
        self.mocker.verify()

    def test_get_share_path(self):
        """Test for get_share_path."""
        share_id = uuid.uuid4()
        name = 'The little pretty share'
        other_name = 'Dorian Grey'
        share = self._create_share_volume(volume_id=share_id, name=name,
                                          other_visible_name=other_name)
        result = get_share_path(share)

        expected = '%s (%s, %s)' % (name, other_name, share_id)
        self.assertEqual(result, expected)

    def test_get_share_path_visible_name_empty(self):
        """Test for get_share_path."""
        share_id = uuid.uuid4()
        name = 'The little pretty share'
        other_name = ''
        share = self._create_share_volume(volume_id=share_id, name=name,
                                          other_visible_name=other_name)
        result = get_share_path(share)

        expected = '%s (%s)' % (name, share_id)
        self.assertEqual(result, expected)
