# -*- coding: utf-8 -*-
#
# Authors: Manuel de la Pena <manuel@canonical.com>
#          Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Platform independent tests for the credentials management."""

from mocker import MockerTestCase, MATCH
from twisted.internet import defer

from ubuntuone.platform.credentials import CredentialsManagementTool


class TestCredentialsManagementTool(MockerTestCase):
    """Test the credentials management tool mocking the proxy."""

    def setUp(self):
        """Set the diff tests."""
        super(TestCredentialsManagementTool, self).setUp()
        self.get_cred_proxy = self.mocker.mock()
        self.proxy = self.mocker.mock()
        self.management_tool = CredentialsManagementTool()
        self.management_tool.get_creds_proxy = self.get_cred_proxy

    def test_find_credentials(self):
        """Test that we do ask for the credentials correctly."""
        signals = ['creds_found', 'creds_not_found', 'creds_erro']
        self.get_cred_proxy()
        self.mocker.result(self.proxy)
        self.proxy.register_to_credentials_found(MATCH(callable))
        self.mocker.result(signals[0])
        self.proxy.register_to_credentials_not_found(MATCH(callable))
        self.mocker.result(signals[1])
        self.proxy.register_to_credentials_error(MATCH(callable))
        self.mocker.result(signals[2])
        self.proxy.find_credentials(reply_handler=MATCH(callable),
                                    error_handler=MATCH(callable))
        self.mocker.replay()
        return self.management_tool.find_credentials()

    def test_clear_credentials(self):
        """Test that we clear the credentials correctly."""
        signals = ['creds_cleared', 'creds_error']
        self.get_cred_proxy()
        self.mocker.result(self.proxy)
        self.proxy.register_to_credentials_cleared(MATCH(callable))
        self.mocker.result(signals[0])
        self.proxy.register_to_credentials_error(MATCH(callable))
        self.mocker.result(signals[1])
        self.proxy.clear_credentials(reply_handler=MATCH(callable),
                                     error_handler=MATCH(callable))
        self.mocker.replay()
        return self.management_tool.clear_credentials()

    def test_store_credentials(self):
        """Test that we store the credentials correctly."""
        token = 'token'
        signals = ['creds_stored','creds_error']
        self.get_cred_proxy()
        self.mocker.result(self.proxy)
        self.proxy.register_to_credentials_stored(MATCH(callable))
        self.mocker.result(signals[0])
        self.proxy.register_to_credentials_error(MATCH(callable))
        self.mocker.result(signals[1])
        self.proxy.store_credentials(token, reply_handler=MATCH(callable),
                                     error_handler=MATCH(callable))
        self.mocker.replay()
        return self.management_tool.store_credentials(token)

    def test_register(self):
        """Test that we correctly register."""
        window_id = 10
        window_id_dict = dict(window_id=str(window_id))
        signals = ['creds_found', 'creds_denied', 'creds_error']
        self.get_cred_proxy()
        self.mocker.result(self.proxy)
        self.proxy.register_to_credentials_found(MATCH(callable))
        self.mocker.result(signals[0])
        self.proxy.register_to_authorization_denied(MATCH(callable))
        self.mocker.result(signals[1])
        self.proxy.register_to_credentials_error(MATCH(callable))
        self.mocker.result(signals[2])
        self.proxy.register(window_id_dict, reply_handler=MATCH(callable),
                            error_handler=MATCH(callable))
        self.mocker.replay()
        return self.management_tool.register(window_id=window_id)

    def test_cleanup_but_no_remove(self):
        """The cleanup method works on handlers that have no remove method."""
        self.management_tool._cleanup_signals = ["test signal 1", "test sig 2"]
        self.management_tool.cleanup(None)


class TestDeferredCredentialsManagementTool(TestCredentialsManagementTool):
    """Test the credentials management tool mocking the proxy in a deferred."""

    def setUp(self):
        """Set the diff tests."""
        super(TestDeferredCredentialsManagementTool, self).setUp()
        real_get_cred_proxy = self.mocker.mock()
        self.get_cred_proxy = lambda: defer.succeed(real_get_cred_proxy())
        self.management_tool = CredentialsManagementTool()
        self.management_tool.get_creds_proxy = self.get_cred_proxy
