# -*- coding: utf-8 -*-
#
# Authors: Manuel de la Pena <manuel@canonical.com>
#          Alejandro J. Cura <alecu@canonical.com>
#
# Copyright 2011 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Ubuntu One credentials management IPC service."""


from ubuntu_sso.main.windows import UbuntuSSOClient
from twisted.internet import defer

from ubuntu_sso.credentials import (
    HELP_TEXT_KEY,
    PING_URL_KEY,
    TC_URL_KEY,
)
from ubuntuone.credentials import (
    CredentialsManagementToolRoot,
    APP_NAME,
    DESCRIPTION,
    NO_OP,
    PING_URL,
    TC_URL,
)
from ubuntuone.credentials import CredentialsError as CredentialsErrorRoot

CredentialsError = CredentialsErrorRoot


class CredentialsManagement(object):
    """Object that manages Ubuntu One credentials."""

    def __init__(self, proxy, *args, **kwargs):
        super(CredentialsManagement, self).__init__(*args, **kwargs)
        self.sso_proxy = proxy

    def find_credentials(self, reply_handler=NO_OP, error_handler=NO_OP):
        """Ask the Ubuntu One credentials."""
        d = self.sso_proxy.find_credentials(APP_NAME, {})
        d.addCallbacks(lambda _:reply_handler(), error_handler)

    def clear_credentials(self, reply_handler=NO_OP, error_handler=NO_OP):
        """Clear the Ubuntu One credentials."""
        self.sso_proxy.clear_credentials(APP_NAME, {},
            reply_handler=reply_handler, error_handler=error_handler)

    def store_credentials(self, credentials,
                          reply_handler=NO_OP, error_handler=NO_OP):
        """Store the token for Ubuntu One application."""
        self.sso_proxy.store_credentials(APP_NAME, credentials,
            reply_handler=reply_handler, error_handler=error_handler)

    def register(self, args, reply_handler=NO_OP, error_handler=NO_OP):
        """Get credentials if found else prompt to register to Ubuntu One."""
        params = {HELP_TEXT_KEY: DESCRIPTION, TC_URL_KEY: TC_URL,
                  PING_URL_KEY: PING_URL}
        params.update(args)
        self.sso_proxy.register(APP_NAME, params,
            reply_handler=reply_handler, error_handler=error_handler)

    def login(self, args, reply_handler=NO_OP, error_handler=NO_OP):
        """Get credentials if found else prompt to login to Ubuntu One."""
        params = {HELP_TEXT_KEY: DESCRIPTION, TC_URL_KEY: TC_URL,
                  PING_URL_KEY: PING_URL}
        params.update(args)
        self.sso_proxy.login(APP_NAME, params,
            reply_handler=reply_handler, error_handler=error_handler)

    def register_to_credentials_stored(self, callback):
        """Register to the CredentialsStored dbus signal."""
        self.sso_proxy.on_credentials_stored_cb = callback
        return self.sso_proxy.on_credentials_stored_cb

    def register_to_credentials_cleared(self, callback):
        """Register to the CredentialsCleared dbus signal."""
        self.sso_proxy.on_credentials_cleared_cb = callback
        return self.sso_proxy.on_credentials_cleared_cb

    def register_to_credentials_found(self, callback):
        """Register to the CredentialsFound dbus signal."""
        pass_credentials = lambda app_name, creds: callback(creds)
        self.sso_proxy.on_credentials_found_cb = pass_credentials
        return self.sso_proxy.on_credentials_found_cb

    def register_to_credentials_not_found(self, callback):
        """Register to the CredentialsFound dbus signal."""
        pass_empty = lambda *args: callback()
        self.sso_proxy.on_credentials_not_found_cb = pass_empty
        return self.sso_proxy.on_credentials_not_found_cb

    def register_to_authorization_denied(self, callback):
        """Register to the AuthorizationDenied dbus signal."""
        self.sso_proxy.on_authorization_denied_cb = callback
        return self.sso_proxy.on_authorization_denied_cb

    def register_to_credentials_error(self, callback):
        """Register to the CredentialsError dbus signal."""
        self.sso_proxy.on_credentials_error_cb = callback
        return self.sso_proxy.on_credentials_error_cb


@defer.inlineCallbacks
def get_creds_proxy():
    """Get the CredentialsManagement dbus proxy."""
    client = UbuntuSSOClient()
    client = yield client.connect()
    yield client.cred_management.register_to_signals()
    defer.returnValue(CredentialsManagement(client.cred_management))


class CredentialsManagementTool(CredentialsManagementToolRoot):
    """Wrapper used to manage the creds of ubuntu one."""

    def __init__(self):
        """Create a new instance."""
        super(CredentialsManagementTool, self).__init__(get_creds_proxy)
