# ubuntuone.platform.windows- windows platform imports
#
# Author: Lucio Torre <lucio.torre@canonical.com>
#
# Copyright 2009 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Windows tools to interact with the os."""

import logging
import os
import sys
import stat

from functools import wraps
from contextlib import contextmanager
from pywintypes import error as PyWinError
from win32api import GetUserName, GetFileAttributes
from win32file import MoveFileExW
from win32com.client import Dispatch
from pywintypes import error
from win32con import FILE_ATTRIBUTE_READONLY
from win32file import (
    MOVEFILE_COPY_ALLOWED,
    MOVEFILE_REPLACE_EXISTING,
    MOVEFILE_WRITE_THROUGH
)
from win32security import (
    ACL,
    ACL_REVISION,
    DACL_SECURITY_INFORMATION,
    CONTAINER_INHERIT_ACE,
    OBJECT_INHERIT_ACE,
    GetFileSecurity,
    LookupAccountName,
    LookupAccountSid,
    SetFileSecurity,
)
from ntsecuritycon import (
    FILE_GENERIC_READ,
    FILE_GENERIC_WRITE,
    FILE_ALL_ACCESS
)

# ugly trick to stop pylint for complaining about
# WindowsError on Linux
if sys.platform != 'win32':
    WindowsError = None


logger = logging.getLogger('ubuntuone.SyncDaemon.VM')
platform = 'win32'

LONG_PATH_PREFIX = '\\\\?\\'
EVERYONE_GROUP = 'Everyone'
ADMINISTRATORS_GROUP = 'Administrators'
# a map that contains the illegal chars and their 'utf8' representation on
# windows so that we can show something to the user
WINDOWS_ILLEGAL_CHARS_MAP = {
    '<':u'\N{ZERO WIDTH SPACE}\N{SINGLE LEFT-POINTING ANGLE QUOTATION MARK}\N{ZERO WIDTH SPACE}',
    '>':u'\N{ZERO WIDTH SPACE}\N{SINGLE RIGHT-POINTING ANGLE QUOTATION MARK}\N{ZERO WIDTH SPACE}',
    ':':u'\N{ZERO WIDTH SPACE}\N{RATIO}\N{ZERO WIDTH SPACE}',
    '"':u'\N{ZERO WIDTH SPACE}\N{DOUBLE PRIME}\N{ZERO WIDTH SPACE}',
    '/':u'\N{ZERO WIDTH SPACE}\N{FRACTION SLASH}\N{ZERO WIDTH SPACE}',
    '|':u'\N{ZERO WIDTH SPACE}\N{DIVIDES}\N{ZERO WIDTH SPACE}',
    '?':u'\N{ZERO WIDTH SPACE}\N{INTERROBANG}\N{ZERO WIDTH SPACE}',
    '*':u'\N{ZERO WIDTH SPACE}\N{SEXTILE}\N{ZERO WIDTH SPACE}'
}
LINUX_ILLEGAL_CHARS_MAP = {
    u'\N{ZERO WIDTH SPACE}\N{SINGLE LEFT-POINTING ANGLE QUOTATION MARK}\N{ZERO WIDTH SPACE}':'<',
    u'\N{ZERO WIDTH SPACE}\N{SINGLE RIGHT-POINTING ANGLE QUOTATION MARK}\N{ZERO WIDTH SPACE}':'>',
    u'\N{ZERO WIDTH SPACE}\N{RATIO}\N{ZERO WIDTH SPACE}':':',
    u'\N{ZERO WIDTH SPACE}\N{DOUBLE PRIME}\N{ZERO WIDTH SPACE}':'"',
    u'\N{ZERO WIDTH SPACE}\N{FRACTION SLASH}\N{ZERO WIDTH SPACE}':'/',
    u'\N{ZERO WIDTH SPACE}\N{SET MINUS}\N{ZERO WIDTH SPACE}':'\\',
    u'\N{ZERO WIDTH SPACE}\N{DIVIDES}\N{ZERO WIDTH SPACE}':'|',
    u'\N{ZERO WIDTH SPACE}\N{INTERROBANG}\N{ZERO WIDTH SPACE}':'?',
    u'\N{ZERO WIDTH SPACE}\N{SEXTILE}\N{ZERO WIDTH SPACE}':'*'
}


def _int_to_bin(n):
    """Convert an int to a bin string of 32 bits."""
    return "".join([str((n >> y) & 1) for y in range(32 - 1, -1, -1)])


def _set_as_long_path(path):
    """Allows to access files longer than 260 chars."""
    if path and not path.startswith(LONG_PATH_PREFIX):
        # return the path with the required prefix added
        return LONG_PATH_PREFIX + path
    return path


def _get_group_sid(group_name):
    """Return the SID for a group with the given name."""
    return LookupAccountName('', group_name)[0]


def _has_read_mask(number):
    """Return if the read flag is present."""
    # get the bin representation of the mask
    binary = _int_to_bin(number)
    # there is actual no documentation of this in MSDN but if bt 28 is set,
    # the mask has full access, more info can be found here:
    # http://www.iu.hio.no/cfengine/docs/cfengine-NT/node47.html
    if binary[28] == '1':
        return True
    # there is no documentation in MSDN about this, but if bit 0 and 3 are true
    # we have the read flag, more info can be found here:
    # http://www.iu.hio.no/cfengine/docs/cfengine-NT/node47.html
    return binary[0] == '1' and binary[3] == '1'


def _check_winerror_123(exception, cb, path, *args, **kwargs):
    """Checks if an exception is a 123 error and executes the cb."""
    if getattr(exception, "winerror", None) == 123 :
        path = remove_windows_illegal_chars(path)
        return cb(path, *args, **kwargs)
    else:
        raise exception


def _contains_chars(path, chars):
    """Returns if one of the chars is in the path."""
    if LONG_PATH_PREFIX in path:
        path = path.replace(LONG_PATH_PREFIX, '')
    return any(c in chars for c in os.path.splitdrive(path)[1])


def _modify_chars(path, should_apply, chars):
    """Remove illegal chars and return a valid windows path."""
    if not should_apply(path):
        return path
    prefix = False
    # do not split the path if you have a long path prefix since the
    # partial paths will be wrong.
    if LONG_PATH_PREFIX in path:
        # remove it and remember to add it
        path = path.replace(LONG_PATH_PREFIX, '')
        prefix = True
    volume, path = os.path.splitdrive(path)
    partial_paths = path.split('\\')
    result = []
    for partial in partial_paths:
        # ensure that the illegal chars are not there
        for illegal_char, replacement in chars.iteritems():
            partial = partial.replace(illegal_char, replacement)
        result.append(partial)
    result = os.path.join(*result)
    if volume:
        result = volume + '\\' + result
    if prefix:
        result = LONG_PATH_PREFIX + result
    return unicode(result)


def is_illegal_path(path):
    """Return if the path is illegal in the system."""
    return _contains_chars(path, WINDOWS_ILLEGAL_CHARS_MAP)


def remove_windows_illegal_chars(path):
    """Remove illegal chars and return a valid windows path."""
    return _modify_chars(path, is_illegal_path, WINDOWS_ILLEGAL_CHARS_MAP)


def is_modified_path(path):
    """Return if the path contains the special utf8 chars."""
    return _contains_chars(path, LINUX_ILLEGAL_CHARS_MAP)


def remove_utf8_chars(path):
    """Remove the special chars from the path and use the basic ones."""
    return _modify_chars(path, is_modified_path, LINUX_ILLEGAL_CHARS_MAP)


def _set_file_attributes(path, groups):
    """Set file attributes using the wind32api."""
    security_descriptor = None
    try:
        security_descriptor = GetFileSecurity(path, DACL_SECURITY_INFORMATION)
    except PyWinError, e:
        logger.exception('Got exception when setting file attributes %s ' +
                         'for path %s', path, groups)
        #check if the issue was due to an illegal path
        if getattr(e, "winerror", None) == 123:
            path = remove_windows_illegal_chars(path)
            return _set_file_attributes(path, groups)
        else:
            raise

    dacl = ACL()
    for group_name in groups:
        # set the attributes of the group only if not null
        if groups[group_name]:
            group_sid = _get_group_sid(group_name)
            dacl.AddAccessAllowedAceEx(ACL_REVISION,
                CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE, groups[group_name],
                group_sid)
    # the dacl has all the info of the diff groups passed in the parameters
    security_descriptor.SetSecurityDescriptorDacl(1, dacl, 0)
    SetFileSecurity(path, DACL_SECURITY_INFORMATION, security_descriptor)


def absparam(paths_indexes=[], paths_names=[]):
    """Ensure that the paths are absolute."""
    def decorator(function):
        """Decorate the function to make sure the paths are absolute."""
        @wraps(function)
        def abspath_wrapper(*args, **kwargs):
            """Set the paths to be absolute."""
            fixed_args = list(args)
            for current_path in paths_indexes:
                fixed_args[current_path] = \
                    os.path.abspath(args[current_path])
            fixed_args = tuple(fixed_args)
            for current_key in paths_names:
                try:
                    kwargs[current_key] = \
                        os.path.abspath(kwargs[current_key])
                except KeyError:
                    logger.warn('Parameter %s could not be made a long path',
                        current_key)
            return function(*fixed_args, **kwargs)
        return abspath_wrapper
    return decorator


def longpath(paths_indexes=[], paths_names=[]):
    """Ensure that the parameters are long paths."""
    def decorator(function):
        """Decorate function so that the given params are long paths.."""
        @wraps(function)
        def long_path_wrapper(*args, **kwargs):
            """Set the paths to be long paths."""
            fixed_args = list(args)
            for current_path in paths_indexes:
                fixed_args[current_path] = \
                    _set_as_long_path(args[current_path])
            fixed_args = tuple(fixed_args)
            for current_key in paths_names:
                try:
                    kwargs[current_key] = \
                        _set_as_long_path(kwargs[current_key])
                except KeyError:
                    logger.warn('Parameter %s could not be made a long path',
                        current_key)
            return function(*fixed_args, **kwargs)
        return long_path_wrapper
    return decorator


def _set_no_rights_helper(path):
    """Set the rights to be none."""
    os.chmod(path, 0o000)
    groups = {}
    _set_file_attributes(path, groups)


@longpath(paths_indexes=[0])
def set_no_rights(path):
    # set the groups to be empty which will remove all the rights of the file
    try:
        _set_no_rights_helper(path)
    except OSError, e:
        logger.exception('Got exception when setting path %s to no rights',
                         path)
        return _check_winerror_123(e, _set_no_rights_helper, path)


@longpath(paths_indexes=[0])
def set_file_readonly(path):
    """Change path permissions to readonly in a file."""
    # we use the win32 api because chmod just sets the readonly flag and
    # we want to have more control over the permissions
    groups = {}
    groups[ADMINISTRATORS_GROUP] = FILE_GENERIC_READ | FILE_GENERIC_WRITE
    groups[GetUserName()] = FILE_GENERIC_READ
    # the above equals more or less to 0444
    _set_file_attributes(path, groups)


@longpath(paths_indexes=[0])
def set_file_readwrite(path):
    """Change path permissions to readwrite in a file."""
    groups = {}
    groups[EVERYONE_GROUP] = FILE_GENERIC_READ
    groups[ADMINISTRATORS_GROUP] = FILE_ALL_ACCESS
    groups[GetUserName()] = FILE_ALL_ACCESS
    # the above equals more or less to 0774
    _set_file_attributes(path, groups)
    try:
        os.chmod(path, stat.S_IWRITE)
    except OSError, e:
        logger.exception('Get exception when trying to set readwrite ' +
                         'path %s', path)
        return _check_winerror_123(e, os.chmod, path, stat.S_IWRITE)


@longpath(paths_indexes=[0])
def set_dir_readonly(path):
    """Change path permissions to readonly in a dir."""
    groups = {}
    groups[ADMINISTRATORS_GROUP] = FILE_GENERIC_READ | FILE_GENERIC_WRITE
    groups[GetUserName()] = FILE_GENERIC_READ
    # the above equals more or less to 0444
    _set_file_attributes(path, groups)


@longpath(paths_indexes=[0])
def set_dir_readwrite(path):
    """Change path permissions to readwrite in a dir."""
    groups = {}
    groups[EVERYONE_GROUP] = FILE_GENERIC_READ
    groups[ADMINISTRATORS_GROUP] = FILE_ALL_ACCESS
    groups[GetUserName()] = FILE_ALL_ACCESS
    # the above equals more or less to 0774
    _set_file_attributes(path, groups)
    # remove the read only flag
    os.chmod(path, stat.S_IWRITE)
    for dirpath, dirs, files in os.walk(path):
        for dir in dirs:
            nested_path = os.path.join(dirpath, dir)
            if not can_write(nested_path):
                set_dir_readwrite(nested_path)
        for file in files:
            nested_path = os.path.join(dirpath, file)
            if not can_write(nested_path):
                set_file_readwrite(nested_path)


@contextmanager
@longpath(paths_indexes=[0])
def allow_writes(path):
    """A very simple context manager to allow witting in RO dirs."""
    # get the old dacl of the file so that we can reset it back when done
    security_descriptor = GetFileSecurity(path, DACL_SECURITY_INFORMATION)
    old_dacl = security_descriptor.GetSecurityDescriptorDacl()
    set_dir_readwrite(path)
    yield
    # set the old dacl back
    security_descriptor.SetSecurityDescriptorDacl(1, old_dacl, 0)
    SetFileSecurity(path, DACL_SECURITY_INFORMATION, security_descriptor)


@longpath(paths_indexes=[0])
def remove_file(path):
    """Remove a file."""
    try:
        os.remove(path)
    except OSError, e:
        logger.exception('Got exception when trying to remove path %s', path)
        return _check_winerror_123(e, os.remove, path)


@longpath(paths_indexes=[0])
def remove_dir(path):
    """Remove a dir."""
    try:
        os.rmdir(path)
    except OSError, e:
        logger.exception('Got exception when trying to remove dir %s', path)
        return _check_winerror_123(e, os.rmdir, path)


@longpath(paths_indexes=[0])
def path_exists(path):
    """Return if the path exists."""
    if is_illegal_path(path):
        path = remove_windows_illegal_chars(path)
    return os.path.exists(path)


@longpath(paths_indexes=[0])
def make_dir(path, recursive=False):
    """Make a dir, optionally creating all the middle ones."""
    if recursive:
        try:
            os.makedirs(path)
        except OSError, e:
            logger.exception('Got exception when trying to make dirs %s', path)
            return _check_winerror_123(e, os.makedirs, path)
    else:
        try:
            os.mkdir(path)
        except OSError, e:
            logger.exception('Got exception when trying to make dirs %s', path)
            return _check_winerror_123(e, os.mkdir, path)


@longpath(paths_indexes=[0])
def open_file(path, mode='r'):
    """Open a file."""
    try:
        return open(path, mode)
    except OSError, e:
        logger.exception('Got exception when trying to open path %s', path)
        return _check_winerror_123(e, open, path, mode)


@absparam(paths_indexes=[0, 1])
@longpath(paths_indexes=[0, 1])
def rename(path_from, path_to):
    """Rename a file or directory."""
    # to ensure the same behaviors as on linux, use the MoveEx function from
    # win32 which will allow to replace the destination path if exists and
    # the user has the rights see:
    # http://msdn.microsoft.com/en-us/library/aa365240(v=vs.85).aspx
    try:
        MoveFileExW(
            path_from, path_to,
            MOVEFILE_COPY_ALLOWED |
            MOVEFILE_REPLACE_EXISTING |
            MOVEFILE_WRITE_THROUGH)
    except PyWinError, e:
        logger.exception('Got exception when trying to rename from ' +
                         '%s to %s', path_from, path_to)
        #check if the issue was due to an illegal path
        if getattr(e, "winerror", None) == 123:
            path_from = remove_windows_illegal_chars(path_from)
            path_to = remove_windows_illegal_chars(path_to)
            try:
                MoveFileExW(
                    path_from, path_to,
                    MOVEFILE_COPY_ALLOWED |
                    MOVEFILE_REPLACE_EXISTING |
                    MOVEFILE_WRITE_THROUGH)
            except PyWinError, e:
                logger.exception('Got exception when trying to rename' +
                                 ' from %s to %s', path_from, path_to)
                raise OSError(str(e))
        else:
            raise OSError(str(e))


@absparam(paths_indexes=[0, 1])
def make_link(target, destination):
    """Create a link from the target in the given destination."""
    # append the correct file type
    if not destination.endswith('.lnk'):
        destination += '.lnk'
    # ensure that the dir containing the link exists
    dirname = os.path.dirname(destination)
    if dirname != '' and not os.path.exists(dirname):
        make_dir(dirname, recursive=True)
    shell = Dispatch('WScript.Shell')
    shortcut = shell.CreateShortCut(destination)
    shortcut.Targetpath = target
    shortcut.save()


@absparam(paths_indexes=[0])
def read_link(path):
    """Read the destination of a link."""
    if not path.endswith('.lnk'):
        path += '.lnk'
    shell = Dispatch('WScript.Shell')
    shortcut = shell.CreateShortCut(path)
    return abspath(shortcut.Targetpath)


@absparam(paths_indexes=[0])
def is_link(path):
    """Returns if a path is a link or not."""
    if not path.endswith('.lnk'):
        path += '.lnk'
    return os.path.exists(path)


@absparam(paths_indexes=[0])
def remove_link(path):
    """Removes a link."""
    if not path.endswith('.lnk'):
        path += '.lnk'
    if os.path.exists(path):
        os.unlink(path)
    

@longpath(paths_indexes=[0])
def listdir(directory):
    """List a directory."""
    # os.listdir combines / and \ in its search which breaks things on
    # windows, we use the win32 api instead.
    if not directory.endswith(os.path.sep):
        directory += os.path.sep
    return os.listdir(directory)

@longpath(paths_indexes=[0])
def access(path):
    """Return if the path is at least readable."""
    # lets consider the access on an illegal path to be a special case
    # since that will only occur in the case where the user created the path
    # for a file to be readable it has to be readable either by the user or
    # by the everyone group
    if is_illegal_path(path):
        path = remove_windows_illegal_chars(path)
    if not os.path.exists(path):
        return False
    security_descriptor = GetFileSecurity(path, DACL_SECURITY_INFORMATION)
    dacl = security_descriptor.GetSecurityDescriptorDacl()
    sids = []
    for index in range(0, dacl.GetAceCount()):
        # add the sid of the ace if it can read to test that we remove
        # the r bitmask and test if the bitmask is the same, if not, it means
        # we could read and removed it.
        ace = dacl.GetAce(index)
        if _has_read_mask(ace[1]):
            sids.append(ace[2])
    accounts = []
    for sid in sids:
        try:
            accounts.append(LookupAccountSid('', sid)[0])
        except error:
            # means that the sid is not linked with a 'visible' account
            # in our case we can ignore this since we are looking for visible
            # users
            continue
    return (GetUserName() in accounts or EVERYONE_GROUP in accounts) and\
            os.access(path, os.R_OK)


@longpath(paths_indexes=[0])
def can_write(path):
    """Return if the path can be written to."""
    file_att = GetFileAttributes(path)
    if file_att and FILE_ATTRIBUTE_READONLY:
        return False
    return True

@longpath(paths_indexes=[0])
def stat_path(path, look_for_link=True):
    """Return stat info about a path."""
    # if the path end with .lnk, that means we are dealing with a link
    # and we should return the stat of the target path
    if path.endswith('.lnk'):
        shell = Dispatch('WScript.Shell')
        shortcut = shell.CreateShortCut(path)
        path = shortcut.Targetpath
    if look_for_link and os.path.exists(path + '.lnk'):
        return stat_path(path + '.lnk')
    try:
        return os.lstat(path)
    except OSError, e:
        logger.exception('Got exception when trying to stat path %s', path)
        return _check_winerror_123(e, stat_path, path, look_for_link)


def move_to_trash(path):
    """Move the file or dir to trash.

    If had any error, or the system can't do it, just remove it.
    """


def set_application_name(app_name):
    """Set the app name."""
    # there is not way to do this on windows. The name will be correct when
    # executed from a bundle .exe otherwise it will be python


def is_root():
    """Return if the user is running as root."""
    # TODO: Do check if we are running as admin
    return False


def expanduser(path):
    """Same as os.path.expanduser."""
    # ensure there are no linux sep
    path = path.replace('/', '\\')
    return LONG_PATH_PREFIX + os.path.expanduser(path)


def validate_path_from_unix(path):
    """Get a unix path and return it so that it can be used."""
    path = path.replace('/', '\\')
    return path


@longpath(paths_indexes=[0])
def get_path_list(path):
    """Return a list with the diff components of the path."""
    path = path.replace(LONG_PATH_PREFIX, '')
    path = path.replace(os.path.expandvars('%HOMEDRIVE%') + os.path.sep, '')
    list = [LONG_PATH_PREFIX + os.path.expandvars('%HOMEDRIVE%'),]
    list.extend(path.split(os.path.sep))
    return list


def abspath(path):
    """Return an abspath."""
    # TODO: do we really need to do this, are the tests broken?
    path = path.replace('/', '\\')
    abs = os.path.abspath(remove_windows_illegal_chars(path))
    if not LONG_PATH_PREFIX in abs:
        abs = LONG_PATH_PREFIX + abs
    return abs


def normpath(path):
    """Normalize path, eliminating double slashes, etc."""
    # the basic normpath does not work correctly with \\?\ which we use
    # as a trick for long paths, lets check it is is there
    if path.startswith(LONG_PATH_PREFIX):
        path = path.replace(LONG_PATH_PREFIX, '')
    return LONG_PATH_PREFIX + os.path.normpath(path)